------------------------------------------------------------------------------------------------
-- Starcraft 2 Model (3ds->M3) Exporter version 0.14
-- by NiNtoxicated (madyavic@gmail.com)
-- Exports M3 models from 3ds max via maxscript
-- Created for 3ds max 2010, but should be backwards compatible
-- Based on reworked code from my World of Warcraft M2 importer/exporter
--
-- Currently exports:
-- - Geometry
-- - Materials
-- - Bounding Box
--
-- Setting up an exportable model:
-- 1. Use one mesh of 'editable_mesh' type only.
-- 2. Apply one material of type Multimaterial, Standard or Starcraft 2 to it.
-- 3. Export!
--
-- Big thanks to: 
-- Teal (starcraft.incgamers.com) - PHP M3 parser
-- Volcore (http://volcore.limbicsoft.com/)  - Helped figure out vertex flags
-- Witchsong (http://code.google.com/p/libm3/) - Awesome M3 library and for help on sequence data
-- Sixen (http://www.sc2mapster.com/) - Providing a good central resource for SC2 tools
-- der_Ton (http://www.doom3world.org/) - Great work on MD5 format which is similar to M3
-- MrMoonKr - Providing a toUpper function to fix 3ds max incompatibility issues
-- ufoZ - One of the original people to reverse engineer the M2 format and provide a good maxscript importer
-- 			from which my importer/exporters were originally based
-- Check out http://code.google.com/p/libm3/ for more indepth detail about the M3 format
-- Head to http://www.sc2mapster.com/ for more mod tools and Starcraft 2 content
--
-- If you use any of this code in your own scripts, please credit
-- where it is due.
--
-- Things to do:
-- 1. Clean up messy code
-- 2. Implement 3dsMax GUI properly
-- 3. Add good animation support
-- 4. Improve code functionality
-- 5. Improve processing times
-- 6. Add more feature support
-- 7. Improve error catching and error support
--
------------------------------------------------------------------------------------------------
--
-- Version History:
-- 0.14 - May 28th 2010
-- Added Alpha Mask map support
-- Added bone support
-- Added baseframe and bindpose support
-- Added attachment support
-- Vastly improved mesh processing 
--
-- 0.09 - May 17th 2010
-- Incompatibility error with old 3ds max versions fixed (thanks to MrMoonKr)
-- Improved material handling
-- Improved code stability
--
-- 0.08 - May 12th 2010
-- Added support for sc2 material and bitmap plugin objects
-- Fixed error with matid's being out of range crashing exporter
-- Improved material code
-- Added more error catching
--
-- 0.07 - May 5th 2010
-- Added some additional UI elements
-- Fixed texture path processing
--
-- 0.06 - May 1st 2010
-- Basic UI added
-- Some basic error catching added
--
-- 0.04 - April 30th 2010
-- Adapted to new MD34 format
-- Basic geometry exporting implemented
-- Bounding sphere exporting added
--
-- 0.01 - April 12th 2010
-- Initial version of M3 export script
--
------------------------------------------------------------------------------------------------

-- ********************
--  SCRIPT DECLARES
-- ********************
-- globals
global head, mwrite, aread, bstream, cbones, m3model
global scrStart, scrEnd -- For script timeStamps
global twrite		= #()

-- Switches
global g_UseInternalPath = false -- use internal path
global g_flipuvy = true -- Flips uv.y coords

-- Misc. variables
global g_mapchannel = 1
global g_bindposeFrame = 0
global g_baseFrame = 1

-- Temp. write file
global texPath = "Assets/Textures/"

-- Function Declares
fn WriteRef bitStream data reftype flags:0 = ()
fn WriteArray bitStream data flags:#uint32 reftype:#null abool:true  = ()

-- *******************
--  M3 STRUCTURES
-- ******************* 
struct M3E_SeqData
(
	tag, ablock
)

struct M3E_Sphere
(
	emin = [0,0,0], emax = [0,0,0], rad = 0, flags = 0,
	
	fn GetBound mesh =
	(
		sph = M3E_Sphere()
		sph.emin = mesh.min
		sph.emax = mesh.max
		local vec3 = mesh.max - mesh.center
		local rad = 0
		for i = 1 to 3 do
		(
			-- set radius to highest x, y or z
			if (vec3[i] > rad) then rad = vec3[i]
		)
		-- increase radius slightly...
		rad *= 1.25
		sph.rad = rad
		
		return sph
	)
)

struct M3E_Matrix
(
	row1 = [1,0,0,0], row2 = [0,1,0,0], row3 = [0,0,1,0], row4 = [0,0,0,1],
	
	fn convObjMatrix objMat =
	(
		m3mat = M3E_Matrix()
		for i = 1 to 4 do
		(
			for j = 1 to 3 do
			(
				case i of
				(
					1: (m3mat.row1[j] = objMat.row1[j])
					2: (m3mat.row2[j] = objMat.row2[j])
					3: (m3mat.row3[j] = objMat.row3[j])
					4: (m3mat.row4[j] = objMat.row4[j])
				)
			)
		)
		
		return m3mat
	)
)

struct M3E_Model
(
	mesh, verts, skin, bones, skinBones = #(), submeshes, usedMats, materials, attachments
)

-- ***********
--  HELPERS
-- ***********
-- Liberally stolen from my M2 script
fn echo msg =
(
	format "%\n" (msg) to:listener
)

-- courtesy of MrMoonKr
-- Converts lower case string to uppercase without needing max 2008
fn M3E_toUpper str =
(
   if str == undefined do return undefined

   upper = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
   lower = "abcdefghijklmnopqrstuvwxyz"

   outstr = copy str

   for i = 1 to outstr.count do
   (
      j = findString lower outstr[i]
      if j != undefined do outstr[i] = upper[j]
   )
   return outstr
)
	
fn LongToString num=
(
	local str = ""
	for i = 1 to 4 do
	(
		str += bit.intAsChar (bit.and num 0xff)
		-- num = bit.shift num -8
		num /= 256
	)
	str
)

fn ShortToFloat &val = -- Short to Float func
(
	if (val > 0) then (val -= 32767) else (val += 32767)
	val /= 32767.0
)

fn M3E_Convert_Time t =
(
	t / (1000 / useFPS)
)

fn M3E_Reset_Globals =
(
	head = mwrite = bstream = m3obj = undefined
	twrite = #()
	gc()
)

-- Writing Funcs
-- Helpers
fn M3E_Convert_Time t =
(
	t * (1000 / useFPS)	
)

fn WriteVec bitStream vec flags bflags:#null =
(
	case flags of
	(
		#v3dshort: (writeShort bitStream vec.x #unsigned; writeShort bitStream vec.y #unsigned; writeShort bitStream vec.z #unsigned;)
		#v3dlong: (writeLong bitStream vec.x #unsigned; writeLong bitStream vec.y #unsigned; writeLong bitStream vec.z #unsigned;)
		#v3dbyte: (writeByte bitStream vec.x #unsigned; writeByte bitStream vec.y #unsigned; writeByte bitStream vec.z #unsigned;)
		#v4d: (writeFloat bitStream vec.x; writeFloat bitStream vec.y; writeFloat bitStream vec.z; writeFloat bitStream vec.w;)
		#v4dshort: (writeShort bitStream vec.x #signed; writeShort bitStream vec.y #signed; writeShort bitStream vec.z #signed; writeShort bitStream vec.w #signed;)
		#v4dbyte: (writeByte bitStream vec.x; writeByte bitStream vec.y; writeByte bitStream vec.z; writeByte bitStream vec.w;)
		#v3d: (writeFloat bitStream vec.x; writeFloat bitStream vec.y; writeFloat bitStream vec.z;)
		#matrix: (WriteVec bitStream vec.row1 #v4d; WriteVec bitStream vec.row2 #v4d; WriteVec bitStream vec.row3 #v4d; WriteVec bitStream vec.row4 #v4d;)
		#sphere: (WriteVec bitStream vec.emin #v3d; WriteVec bitStream vec.emax #v3d; WriteFloat bitStream vec.rad;)
	)
	if (bflags == #MD34) then WriteLong bitStream 0x00 #unsigned
)

fn WriteBlank bitStream count flags:#buffer ret:false =
(
	local bm = ftell bitStream
	local bcount = 0 -- buffer count

	if (flags != #blank) then
	(
		bcount = (16 - (mod count 16)) -- calculate buffer bytes
		if (mod count 16 == 0) then bcount -= 16 -- so we don't get rows of 16 buffer bytes
	)
	
	case flags of
	(
		#buffer: 
		(
			count = 0 -- don't let it write null data
		)
	)
	
	for i = 1 to (count as integer) do (WriteByte bitStream 0) -- write blank bytes
	for i = 1 to (bcount as integer) do (WriteByte bitStream 0xAA #unsigned) -- write buffer bytes
	if (ret != false) then fseek bitStream bm #seek_set
)

fn FloatToShort &val = -- Float to Short func
(
	if (val > 0) then (val -= 1) else (val += 1)
	val *= 32767.0
)

fn M3E_getSize type flags:0 =
(
	local size
	case type of
	(
		#REF: size = 0x0C
		#AREF: size = 0x20
		#MD33: size = 0x14
		#MD34: size = 0x18
		#MODL:
		(
			case flags of
			(
				23: size = 0x310
			)
		)
		#SEQS: size = 0x60
		#STC_: size = 0xCC
		#SDEV: size = M3E_getSize #AREF
		#EVNT: size = 0x68
		#STG_: size = 0x18
		#STS_: size = 0x1C
		#BONE: size = 0xA0
		#VERT: size = 0x20
		#U16_: size = 0x02
		#U32_: size = 0x04
		#I32_: size = 0x04
		#CHAR: size = 0x01
		#DIV_: size = 0x34
		#REGN: size = 0x24
		#BAT_: size = 0x0E
		#MSEC: size = 0x48
		#ATT_: size = 0x14
		#MATM: size = 0x08
		#MAT_: size = 0x10C
		#LAYR: size = 0x164
		#IREF: size = 0x40
	)
	
	return size
)

fn M3E_Name name flt ext:undefined =
(
	-- Filter name from path
	local mn = filterString name flt
	mn = mn[mn.count]
	local mext = substring mn (mn.count-3) (4)
	-- Remove extension if specified
	if (ext != undefined) then
	(
		if (matchPattern mext pattern:ext ignoreCase:true) then mn = substring mn 1 (mn.count-4)
	)
	return mn
)

-- ************************
-- M3 FUNCTIONS
-- ************************ 
fn M3E_Open fname flags =
(
	step = "Accessing Model File"
	
	echo ("Writing "+fname+"...")
	local bitStream = fopen fname flags
	if bitStream==undefined then 
	(
		echo "File not found!"
		throw "File not found"
	)
	else return bitStream
)

fn M3E_Close bitStream =
(
	step = "Close"
	fclose bitStream
)

struct M3E_Tag -- Tag data
(
	dataID, ofsData, nData, flags,
	
	fn Write bitStream tag =
	(
		-- head to end of file
		fseek bitStream 0 #seek_end
		
		-- Fix Header info
		head.ofsTag = ftell bitStream
		head.nTag = tag.count
		
		-- write tags
		for i = 1 to tag.count do
		(
			t = tag[i]
			writeString bitStream t.dataID -- Write File ID
			fseek bitStream -1 #seek_cur
			writeLong bitStream t.ofsData #unsigned
			writeLong bitStream t.nData #unsigned
			writeLong bitStream t.flags #unsigned
		)
	),
	fn TagAppend bitStream nameID nData tagar flags:0 =
	(
		tag = M3E_Tag()
		local strID = nameID as string
		local tagID = "BLNK" -- Needs to be 4 chars for modification
		for i = 1 to 4 do
		(
			k = 5 - i
			tagID[i] = strID[k]
		)
		tag.dataID = M3E_toUpper tagID -- Uppercase
		tag.ofsData = ftell bitStream
		tag.nData = nData
		tag.flags = flags
		
		append tagar tag
	)
)

struct M3E_Ref
(
	entries, refid, flags = 0, data = #(),
	
	fn Write bitStream ref =
	(
		writeLong bitStream ref.entries #unsigned
		writeLong bitStream ref.refid #unsigned
		WriteLong bitStream ref.flags #unsigned
	)
)

struct M3E_AnimRef -- Animation Reference
(
	flags, animflags, animid,
	
	fn Write bitStream ar =
	(
		writeShort bitStream ar.flags #unsigned
		writeShort bitStream ar.animflags #unsigned
		writeLong bitStream ar.animid #unsigned
	),
	fn Default =
	(
		ar = M3E_AnimRef()
		ar.flags = 0
		ar.animflags = 0
		-- random uint32
		ar.animid = Random 1 100000000000
		
		return ar
	)
)

struct M3E_Header
(
	fileID, ofsTag, nTag, mref = M3E_Ref(),
	
	fn Write bitStream head =
	(
		step = "Write Header"

		fseek bstream 0 #seek_set -- move to beginning of file
		
		writeString bitStream head.fileID -- Write File ID
		fseek bitStream -1 #seek_cur -- writeString writes null terminator...so have to move cursor back one space
		writeLong bitStream head.ofsTag #unsigned
		writeLong bitStream head.nTag #unsigned
		M3E_Ref.Write bitStream head.mref
	),
	fn Default =
	(
		h = M3E_Header()
		-- Fixed file magic ID
		h.fileID = "43DM"
		-- Blank these for now, fix em up during tag writing
		h.ofsTag = 0
		h.nTag = 0
		-- Use fixed MODL ref...
		h.mref.entries = 1
		h.mref.refid = 1
		
		return h
	)
)

struct M3E_MODL
(
	name,  -- max model name
	versID,
	SEQS, STC, STG, vec1, STS,
	bones, nSkinBones,
	vflags, verts, DIV, bonelu,
    ext1, attach, attachlu,
	matm, mat,
	IREF,
	
	fn Write bitStream modl vers = 
	(
		-- goto modl offset first
		fseek bitStream twrite[2].ofsData #seek_set
		
		-- Write MODL chunk
		WriteRef bitStream modl.name #CHAR
		WriteLong bitStream modl.versID #unsigned
		WriteRef bitStream modl.SEQS #SEQS flags:1
		WriteRef bitStream modl.STC #STC_ flags:4
		WriteRef bitStream modl.STG #STG_
		WriteVec bitStream modl.vec1 #v3d bflags:#MD34
		WriteRef bitStream modl.STS #STS_
		WriteRef bitStream modl.bones #BONE flags:1
		WriteLong bitStream modl.nSkinBones #unsigned
		WriteLong bitStream modl.vflags #unsigned
		WriteRef bitStream modl.verts #VERT
		WriteRef bitStream modl.DIV #DIV_ flags:2
		WriteRef bitStream modl.bonelu #U16_ subtype:#uint16
		WriteVec bitStream modl.ext1 #sphere bflags:#MD34
		WriteBlank bitStream 0x3C flags:#blank
		WriteRef bitStream modl.attach #ATT_ flags:1
		WriteRef bitStream modl.attachlu #U16_ subtype:#int16
		WriteBlank bitStream 0x30 flags:#blank
		WriteRef bitStream modl.MATM #MATM
		WriteRef bitStream modl.MAT #MAT_ flags:15
		WriteBlank bitStream 0xFC flags:#blank
		WriteRef bitStream modl.iref #IREF
	)
)

struct M3E_SEQS
(
	d1 = #(-1, -1), name, d2 = 0, maxframes, flags, d3 = #(0, 100, 1, 1, 100, 0), bndSphere = M3E_Sphere(), d4 = #(0,0),
	cstart, cend, -- Max variables
	
	fn Write bitStream ad =
	(
		WriteArray bitStream ad.d1 flags:#int32
		WriteRef bitStream ad.name #CHAR
		WriteLong bitStream ad.d2 #unsigned
		WriteLong bitStream ad.maxframes #unsigned
		WriteLong bitStream ad.flags #unsigned
		WriteArray bitStream ad.d3 flags:#uint32
		WriteVec bitStream ad.bndSphere #sphere
		WriteArray bitStream ad.d4 flags:#int32
	),
	fn Default m3obj =
	(
		a = M3E_SEQS()
		
		a.name = "Stand"
		a.maxframes = 10000
		a.flags = 0
		a.bndSphere = M3E_Sphere.GetBound m3obj.mesh
		
		return a
	)
)

struct M3E_Event
(
	name = "Evt_SeqEnd", d1 = -1, s1 = -1, s2 = 0, mat1 = M3E_Matrix(), d2 = #(4,0,0,0),
	
	fn Write bitStream evnt =
	(
		WriteRef bitStream evnt.name #CHAR
		WriteLong bitStream evnt.d1 #signed
		WriteShort bitStream evnt.s1 #signed
		WriteShort bitStream evnt.s2 #unsigned
		WriteVec bitStream evnt.mat1 #matrix bflags:#null
		WriteArray bitStream evnt.d2 flags:#uint32
	)
)

struct M3E_AnimBlock -- Sequence Data
(
	frames, flags, fend, keys,
	
	fn Write bitStream ablock sdtype =
	(
		WriteRef bitStream ablock.frames #I32_ subtype:#uint32
		WriteLong bitStream ablock.flags #unsigned
		WriteLong bitStream ablock.fend #unsigned
		WriteRef bitStream ablock.keys sdtype
	),
	fn DefaultEvnt =
	(
		a = M3E_AnimBlock()
		a.frames = #(9999)
		a.flags = 1
		a.fend = 10000
		a.keys = #(M3E_Event())
		
		return a
	)
)

struct M3E_Animoffs 
(
	aind = 0, sdind = 0,
	
	fn Write bitStream aoffs =
	(
		WriteShort bitStream aoffs.aind #unsigned
		WriteShort bitStream aoffs.sdind #unsigned
	)
)

struct M3E_STC --Sequence Transformations Collection (STC)
(
	name, d1, seqlu, stglu, animid, animref = M3E_Animoffs(), d2, sdata = #(),
	
	fn Write bitStream stc =
	(
		WriteRef bitStream stc.name #CHAR
		WriteLong bitStream stc.d1 #unsigned
		WriteShort bitStream stc.seqlu #unsigned
		WriteShort bitStream stc.stglu #unsigned
		WriteRef bitStream stc.animid #U32_ subtype:#uint32
		WriteRef bitStream stc.animref #U32_ subtype:#animind
		WriteLong bitStream stc.d2 #unsigned
		WriteRef bitStream stc.sdata[1] #SDEV
		for i = 2 to 13 do WriteRef bitStream stc.sdata[i] #null
	),
	fn Default =
	(
		s = M3E_STC()
		s.name = "Stand_full"
		s.d1 = 0
		s.seqlu = 0
		s.stglu = 0
		s.animid = #(Random 1 100000000000) -- must be array
		s.animref = #(M3E_Animoffs())
		s.d2 = 0
		-- initialize array
		s.sdata[13] = 0
		s.sdata[1] = #(M3E_AnimBlock.DefaultEvnt())
		for i = 2 to 13 do s.sdata[i] = #null
		
		return s
	)
)

struct M3E_STG
(
	name = "Stand", stcind = #(0),
	
	fn Write bitStream stg =
	(
		WriteRef bitStream stg.name #CHAR
		WriteRef bitStream stg.stcind #U32_ subtype:#uint32
	)
)

struct M3E_STS
(
	animid, d1 = #(-1,-1,-1), s1 = -1, s2 = 0,
	
	fn Write bitStream sts =
	(
		WriteRef bitStream sts.animid #U32_ subtype:#uint32
		WriteArray bitStream sts.d1 flags:#int32
		WriteShort bitStream sts.s1 #signed
		WriteShort bitStream sts.s2 #unsigned
	),
	fn Default modl =
	(
		s = M3E_STS()
		
		s.animid = #(modl.stc[1].animid[1])
		
		return s
	)
)

struct M3E_Bone
(
	d1, name, flags, parent, s1, 
	transar, pos1, pos2, d2,  
	rotar, rot1, rot2, d3, 
	scalear, scale1, scale2, 
	d4, ar1, unk1, unk2, d5,
	boneIndex, parentptr, bindmat, invbindmat, maxObj, iref,
	
	fn Write bitStream bone =
	(
		WriteLong bitStream bone.d1 #signed
		WriteRef bitStream bone.name #CHAR
		WriteLong bitStream bone.flags #unsigned
		WriteShort bitStream bone.parent #signed
		WriteShort bitStream bone.s1 #unsigned
		M3E_AnimRef.Write bitStream bone.transar
		WriteVec bitStream bone.pos1 #v3d
		WriteVec bitStream bone.pos2 #v3d
		WriteLong bitStream bone.d2 #unsigned
		M3E_AnimRef.Write bitStream bone.rotar
		WriteVec bitStream bone.rot1 #v4d
		WriteVec bitStream bone.rot2 #v4d
		WriteLong bitStream bone.d3 #uint32
		M3E_AnimRef.Write bitStream bone.scalear
		WriteVec bitStream bone.scale1 #v3d
		WriteVec bitStream bone.scale2 #v3d
		WriteLong bitStream bone.d4 #unsigned
		M3E_AnimRef.Write bitStream bone.ar1
		WriteLong bitStream bone.unk1 #unsigned
		WriteLong bitStream bone.unk2 #unsigned
		WriteLong bitStream bone.d5 #unsigned
	),
	fn Default =
	(
		b = M3E_Bone()
		
		b.d1 = -1
		b.name = "Main"
		b.flags = 2048
		b.parent = -1
		b.s1 = 0
		b.transar = M3E_AnimRef.Default()
		b.pos1 = [0,0,0]
		b.pos2 = [0,0,0]
		b.d2 = 0
		b.rotar = M3E_AnimRef.Default()
		b.rot1 = [0,0,0,1]
		b.rot2 = [0,0,0,1]
		b.d3 = 0
		b.scalear = M3E_AnimRef.Default()
		b.scale1 = [1,1,1]
		b.scale2 = [1,1,1]
		b.d4 = 0
		b.ar1 = M3E_AnimRef.Default()
		b.unk1 = 1
		b.unk2 = 1
		b.d5 = 0
		
		b.iref = M3E_Matrix()
		
		return b
	)		
)

struct M3E_Vertex
(
	pos = [0,0,0], bw = #(0,0,0,0), bi = #(0,0,0,0), normal = #(0,0,0,0), uv = #(0,0), d1 = 0, d2 = 0,
	faceid, faceref = #(), origvertindx, origtvertindx, -- for script purposes only
	
	fn Write bitStream vert =
	(
		WriteVec bitStream vert.pos #v3d
		WriteArray bitStream vert.bw flags:#uint8
		WriteArray bitStream vert.bi flags:#uint8
		WriteArray bitStream vert.normal flags:#uint8
		WriteArray bitStream vert.uv flags:#int16
		--WriteLong bitStream vert.d1 #unsigned
		WriteLong bitStream vert.d2 #unsigned
	)
)

struct M3E_REGN
(
	s1 = #(0,0), indvert, nvert, 
	indface, nface, 
	bcount = 1, indbonelu = 0, nbonelu = 1,
	s2 = 0, b1 = #(1,1), s3 = 0,
	verts = #(), faces = #(), vertsList = #(), skinBones = #(), -- maxscript holders

	fn Write bitStream regn =
	(
		WriteArray bitStream regn.s1 flags:#uint32
		WriteLong bitStream regn.indvert #unsigned
		WriteLong bitStream regn.nvert #unsigned
		WriteLong bitStream regn.indface #unsigned
		WriteLong bitStream regn.nface #unsigned
		WriteShort bitStream regn.bcount #unsigned
		WriteShort bitStream regn.indbonelu #unsigned
		WriteShort bitStream regn.nbonelu #unsigned
		WriteShort bitStream regn.s2 #unsigned
		WriteArray bitStream regn.b1 flags:#uint8
		WriteShort bitStream regn.s3 #unsigned
	)
)

struct M3E_BAT
(
	d1 = 0, subid, d2 = 0, matid, s1 = -1,
	
	fn Write bitStream bat =
	(
		WriteLong bitStream bat.d1 #unsigned
		WriteShort bitStream bat.subid #unsigned
		WriteLong bitStream bat.d2 #unsigned
		WriteShort bitStream bat.matid #unsigned
		WriteShort bitstream bat.s1 #signed
	)
)

struct M3E_MSEC
(
	d1 = 0, ar1 = M3E_AnimRef.Default(), ext1 = M3E_Sphere(), d8 = #(0,0,0,0,0,0,0,0),
	
	fn Write bitStream msec =
	(
		WriteLong bitStream msec.d1 #unsigned
		M3E_AnimRef.Write bitStream msec.ar1
		WriteVec bitStream msec.ext1 #sphere
		WriteArray bitStream msec.d8 flags:#uint32
	)
)

struct M3E_DIV -- Model divisions
(
	faces = #(), regn = #(), bat = #(), msec = #(), d1 = 1,
	
	fn Write bitStream div =
	(
		WriteRef bitStream div.faces #U16_ subtype:#uint16
		WriteRef bitStream div.regn #REGN flags:3
		WriteRef bitStream div.bat #BAT_ flags:1
		WriteRef bitStream div.msec #MSEC flags:1
		WriteLong bitStream div.d1 #unsigned
	)
)

struct M3E_Subref -- Submesh refrences
(
	subid, matid
)

struct M3E_Attachment
(
	flags = -1, name, boneParent,
	
	fn Write bitStream m3attach =
	(
		WriteLong bitStream m3attach.flags #signed
		WriteRef bitStream m3attach.name #CHAR
		WriteLong bitStream m3attach.boneParent #unsigned
	)
)

struct M3E_MATM
(
	nmat = 1, matind,
	
	fn Write bitStream matm =
	(
		WriteLong bitStream matm.nmat #unsigned
		WriteLong bitStream matm.matind #unsigned
	),
	fn Read mats =
	(
		marray = #()
		for i = 1 to mats.count do
		(
			m = M3E_MATM()
			m.matind = mats[i] - 1
			
			append marray m
		)
		
		return marray
	)
)

struct M3E_LAYR --more work to be done at a later point
(
	name, alphaFlags = 0, flags = 236, f1 = 1, f2 = 1, 
	
	fn Write bitStream layr =
	(
		WriteLong bitStream 0 #unsigned
		WriteRef bitStream layr.name #CHAR
		WriteBlank bitStream 0x14 flags:#blank
		WriteLong bitStream layr.flags #unsigned
		WriteBlank bitStream 0x04 flags:#blank
		WriteLong bitStream layr.alphaFlags #unsigned
		WriteBlank bitStream 0x08 flags:#blank
		WriteFloat bitStream layr.f1
		WriteFloat bitStream layr.f2
		WriteBlank bitStream 0x1C flags:#blank
		WriteLong bitStream -1 #signed
		WriteBlank bitStream 0xEC flags:#blank
		WriteLong bitStream -1 #signed
		WriteBlank bitStream 0x14 flags:#blank
	)
)

struct M3E_Mat
(
	name, flags = 0, 
	blendMode = 0, priority = 0, 
	spec = 0, alphaMaskEnable = 0, spec_mult = 1, emis_mult = 1, cutoutThresh = 0,
	layerBlendType = 2, emisBlendType = 2,
	layrs = #(),
	
	fn Write bitStream mat =
	(
		WriteRef bitStream mat.name #CHAR
		WriteBlank bitStream 0x04 flags:#blank
		WriteLong bitStream mat.flags #unsigned
		WriteLong bitStream mat.blendMode #unsigned
		WriteLong bitStream mat.priority #unsigned
		WriteBlank bitStream 0x04 flags:#blank
		WriteFloat bitStream mat.spec
		WriteBlank bitStream 0x04 flags:#blank
		WriteLong bitStream mat.alphaMaskEnable #unsigned
		WriteFloat bitStream mat.spec_mult
		WriteFloat bitStream mat.emis_mult
		-- Layer Refs
		for i = 1 to 13 do WriteRef bitStream mat.layrs[i] #LAYR flags:22
			
		WriteLong bitStream mat.cutoutThresh #unsigned
		WriteLong bitStream mat.layerBlendType #unsigned
		WriteLong bitStream mat.emisBlendType #unsigned
		
		WriteBlank bitStream 0x30 flags:#blank
	)
)

-- ****************
--  EXPORT CODE
-- ****************
---------------------
-- SCENE CODE --
---------------------
fn M3E_getObjects objCategory objClass =
(
	echo ("Getting "+objClass+" objects...")
	objs = #()
	Select objCategory
	
	local sel = getCurrentSelection() -- store all objects
	
	for i = 1 to sel.count do
	(
		curObj = sel[i]
		local curObjClass = classOf curObj as string
		
		if (curObjClass != objClass or curObj.parent == undefined) then
		(
			deselect curObj
		)
		else
		(
			append objs curObj
		)
	)
	
	if (objs.count != 0) then return objs else return undefined
)

fn M3E_buildAttachments objPoints m3bones =
(
	local m3attachs = #()
	if (objPoints != undefined) then
	(
		echo "Building attachments..."
		for i = 1 to objPoints.count do
		(
			local a = objPoints[i]
			local m3attach = M3E_Attachment()
			
			m3attach.name = a.name
			for j = 1 to m3bones.count do
			(
				b = m3bones[j]
				if (a.parent.name == b.name) then m3attach.boneParent = j - 1
			)
			
			if (m3attach.boneParent != undefined) then append m3attachs m3attach
		)
		
		fn comparePar a1 a2 = if (a1.boneParent > a2.boneParent) then return 1 else if (a1.boneParent < a2.boneParent) then return -1 else return 0
		qsort m3attachs comparePar
	)
	
	if (m3attachs.count > 0) then
	(
		return m3attachs
	)
	else return undefined
)

fn M3E_getMesh =
(
	echo "Getting mesh..."
	cm = #()
	Select geometry
	local sel = getCurrentSelection() -- store Mesh!
	
	max modify mode
	set coordsys world
	
	-- Mesh sort, deselect unsuitable meshes
	for i = 1 to sel.count do
	(
		m = sel[i]
		meshClass = classOf m.baseObject as string
		if (meshClass != "Editable_mesh") then
		(
			--echo ("#Warning! "+w_mesh as string+" is not an Editable_mesh class!")
			--echo ("Class of object: "+classof w_mesh.baseObject as string)
			deselect m
		)
		else
		(
			if (m.isHidden == true) or (m.isFrozen == true) then -- do not export hidden/frozen geometry, will make this a selectable option!
			(
				deselect m
			)
			else
			(
				-- append to mesh list
				append cm m
			)
		)
	)

	if cm.count < 1 then
	(
		messagebox "No Editable Meshes Found!"
		throw "No meshes found"
	)
	else if cm.count > 1 then
	(
		messagebox "Only one Editable_Mesh allowed!"
		throw "Greater than one editable_mesh"		
	)
	
	local expMesh = cm[1]
	
	return expMesh
)

fn M3E_getUsedMats mesh =
(
	echo "Gathering used materials..."
	local mats = #()
	-- error catching
	if (mesh.material == undefined) then 
	(
		messagebox "No material assigned to your mesh, you must assign your mesh a material of type Multimaterial, Standard or Starcraft 2!"
		throw "No material assigned to mesh"				
	)
	
	local parentmatclass = classOf mesh.material as string
	if (parentmatclass != "Multimaterial" and parentmatclass != "Standardmaterial" and parentmatclass != "sc2material") then
	(
		messagebox "Mesh material type must be Multimaterial, Standard or Starcraft 2"
		throw "Incompatible mesh material type"
	)
	
	-- cycle through faces, gather data
	for k = 1 to mesh.numfaces do
	(
		-- #MATERIALS/SUBMESHES
		-- Gather Used Materials
		local matid = getfacematid mesh k
		local found = false
		local j = 1
		
		if (parentmatclass == "Multimaterial") then
		(
			-- prevent error for out of range matid's
			if (matid > mesh.material.count) then 
			(
				local subid = (mod matid mesh.material.count) as integer
				if (subid == 0) then subid = mesh.material.count
				matid = subid
			)
		)
		else
		(
			matid = 1
		)
		
		while (j <= mats.count and found == false) do
		(
			if mats[j] == matid then 
			(
				found = true
			)
			j+=1
		)
		if found == false then
		(
			append mats matid
		)
	)
	
	-- order the mats
	sort mats
	
	return mats
)

fn M3E_getSkin mesh =
(
	format "Checking for skin..." to:listener
	local cskin
	-- Find Skin class
	for i = 1 to mesh.modifiers.count do
	(
		local meshMod = mesh.modifiers[i]
		local modClass = classOf meshMod as string
		
		if (modClass == "Skin") then
		(
			cskin = meshMod
		)
	)
	
	if (cskin != undefined) then
	(
		echo "found!"
		return cskin
	)
	else
	(
		echo "#WARNING: No skin found! Bones will not be exported"
		return undefined
	)
)

fn M3E_Bone_ParentDepth bind mbones boneDepth =
(
	-- For organising bones based on depth
	local boneParent = mbones[bind].parent + 1
	if boneParent != 0 then 
	(
		boneDepth[boneParent].y += 1
		M3E_Bone_ParentDepth boneParent mbones boneDepth
	)
)

fn M3E_Bone_ChildDepth bind mbones =
(
	-- For organising bones based on depth
	local boneParent = mbones[bind].parent + 1
	if boneParent == 0 then 0
	else return ( 1 + M3E_Bone_ChildDepth boneParent mbones  )
)

fn M3E_Get_Bone_Depth linkBones =
(
	-- sort bones by depth
	bDepth = #()
	bDepth[linkBones.count] = 0
	
	-- do child depth
	for i=1 to linkBones.count do
	(
		bonerec = [(M3E_Bone_ChildDepth i linkBones), 0, i]
		bDepth[i] = bonerec
	)
	
	-- do parent depth
	for i = 1 to linkBones.count do
	(
		M3E_Bone_ParentDepth i linkBones bDepth
	)

	fn compfn a b = 
	( 
		-- order by first criteria
		if a.y < b.y then 
		(
			return 1
		)
		else if a.y > b.y then 
		(
			return -1
		)
		else if a.y == b.y then
		(
			-- order by second criteria
			if a.x < b.x then
			(
				return 1
			)
			else if a.x > b.x then
			(
				return -1
			)
			else return 0
		)
	)
	
	qsort bDepth compfn

	return bDepth
)

fn M3E_buildBoneChain obj newArray = 
(
	local boneExists = false
	for i = 1 to newArray.count do
	(
		if (obj == newArray[i]) then boneExists = true
	)
	
	if (boneExists == false) then
	(
		if ((obj.baseObject as string) == "Bone") then
		(
			append newArray obj
			for i = 1 to obj.children.count do
			(
				M3E_buildBoneChain obj.children[i] newArray
			)
		)
	)
)

fn M3E_buildBones mbones =
(
	-- set temp. parents
	for i = 1 to mbones.count do
	(
		bchild = mbones[i]
		if bchild.parentptr != undefined then
		(
			for j = 1 to mbones.count do
			(
				bpar = mbones[j]
				if (bchild.parentptr == bpar.maxObj) then bchild.parent = (j - 1)
			)
		)
	)
	
	-- order by depth (real bone list)
	local boneDepth = M3E_Get_Bone_Depth mbones
	
	-- build bone chains
	local boneChains = #()
	--newBoneIndex[mbones.count] = 0
	for i = 1 to mbones.count do
	(
		-- build chains based on depth
		local j = boneDepth[i].z
		M3E_buildBoneChain mbones[j].maxObj boneChains
	)
	
	-- build proper index
	local boneIndex = #()
	boneIndex[mbones.count] = 0
	for i = 1 to boneChains.count do
	(
		for j = 1 to mbones.count do
		(
			if (boneChains[i] == mbones[j].maxObj) then
			(
				boneIndex[i] = mbones[j]
			)
		)
	)
	
	/*
	for i = 1 to boneChains.count do
	(
		echo ("boneChains["+i as string+"]: "+boneChains[i] as string)
	)
	*/
	
	-- Setup real bone list
	for i = 1 to boneIndex.count do
	(
		bchild = boneIndex[i]
		if bchild.parentptr != undefined then
		(
			for j = 1 to mbones.count do
			(
				bpar = boneIndex[j]
				if (bchild.parentptr == bpar.maxObj) then 
				(
					bchild.parent = (j - 1)
					local wbindmat = bchild.bindmat * bpar.invbindmat
					bchild.rot1 = inverse wbindmat.rotationpart
					bchild.pos1 = wbindmat.translation
					bchild.scale1 = wbindmat.scalepart
				)
			)
		)
		else
		(
			bchild.rot1 = inverse bchild.bindmat.rotationpart
			bchild.pos1 = bchild.bindmat.translation
			bchild.scale1 = bchild.bindmat.scalepart
		)
	)
	
	-- replace old array with new ordered by depth array
	for i = 1 to mbones.count do
	(
		mbones[i] = boneIndex[i]
	)
)

fn M3E_getBones skin =
(
	echo "Getting bones..."
	local mbones = #()
	
	if (skin != undefined) then
	(
		
		max modify mode
		modPanel.setCurrentObject skin
		
		local nBones = skinops.getnumberbones skin
		for i = 1 to nBones do
		(
			local m3Bone = M3E_Bone.Default()
			
			local maxBoneName = skinops.getbonename skin i 0
			local maxBone = getnodebyname maxBoneName
			m3bone.maxObj = maxBone
			m3bone.name = skinops.getbonename skin i 1
			if (maxBone.parent != undefined) then m3bone.parentptr = maxBone.parent
			
			-- get IREF transform
			at time g_bindposeFrame
			(
				m3bone.iref = M3E_Matrix.convObjMatrix (inverse maxBone.transform)
			)
			
			-- get bindpose transform
			at time g_baseFrame
			(
				m3bone.bindmat = maxBone.transform
			)
			
			m3bone.invbindmat = inverse m3bone.bindmat

			append mbones m3bone
		)
		-- Construct bone array
		M3E_buildBones mbones
	)
	else
	(
		-- Default single bone when no skin present
		m3bone = M3E_Bone.Default()
		append mbones m3bone
	)
	
	return mbones
)

fn M3E_buildFaces obj triindx regn =
(
	-- annoyingly complicated!
	vertindices = getface obj triindx
	if (meshop.getMapSupport obj 1) then
		tvertindices = meshop.getmapface obj g_mapchannel triindx
	else
		tvertindices = [0,0,0]
	vertindices = #(vertindices.x as integer, vertindices.y as integer, vertindices.z as integer)
	tvertindices = #(tvertindices.x as integer, tvertindices.y as integer, tvertindices.z as integer)
	m3vertindices = #()
	for i=1 to 3 do
	(
		vind = vertindices[i]
		tvind = tvertindices[i]
		-- findItem vastly more efficient over old function method
		local index = findItem regn.vertsList [vind, tvind]
		if (index == 0) then
		(
			v = M3E_Vertex()
			v.pos 	= getVert obj vind
			v.normal	= getNormal obj vind
			if (meshop.getMapSupport obj g_mapchannel) then
				v.uv = meshop.getmapvert obj g_mapchannel tvind
			else uv = [0,0,0]
			v.origvertindx = vind
			v.origtvertindx = tvind
			append regn.verts v
			append regn.vertsList [vind, tvind]
			m3vertindices[i] = regn.verts.count
		)
		else
		(
			m3vertindices[i] = index
		)
	)
	
	for i = 1 to 3 do
	(
		append regn.faces (m3vertindices[i] - 1)
	)
)

fn M3E_buildDIV m3obj =
(
	-- run through submeshes
	local div = m3obj.submeshes
	for i = 1 to div.regn.count do
	(
		regn = div.regn[i]
		
		-- Order submesh verts and faces
		regn.nvert = regn.verts.count
		regn.nface = regn.faces.count
		if (i > 1) then
		(
			prevregn = div.regn[i-1]
			regn.indvert = prevregn.indvert + prevregn.nvert
			regn.indface = prevregn.indface + prevregn.nface
		)
		else
		(
			regn.indvert = 0
			regn.indface = 0
		)
		
		-- handle skinned bones if skin present
		if (m3obj.skin != undefined) then
		(
			regn.bcount = regn.skinBones.count
			if (i > 1) then 
			(
				local prevregn = div.regn[i - 1]
				regn.indbonelu = prevregn.indbonelu + prevregn.bcount
			)
			else
			(
				regn.indbonelu = 0
			)
		)
		else
		(
			regn.bcount = 1
			regn.indbonelu = 0
		)
		
		regn.nbonelu = regn.bcount
		
		-- build single faces list
		join div.faces regn.faces
	)
	
	div.msec = #(M3E_MSEC())
	
	-- assign new div back to main container
	m3obj.submeshes = div
)

fn M3E_getSubmesh m3obj =
(
	echo "Getting submeshes..."
	div = M3E_DIV()
	local mesh = m3obj.mesh
	local mats = m3obj.usedMats
	
	for i = 1 to mats.count do
	(
		-- Create Submeshes
		r = M3E_REGN()
		b = M3E_BAT()
		b.subid = i - 1
		b.matid = mats[i] - 1

		append div.regn r
		append div.bat b
	)

	-- collect faces into the submeshes, according to the material the faces use
	matclass = classOf mesh.material as string
	max modify mode
	set time g_bindposeFrame
	--modPanel.setCurrentObject m3obj.mesh
	
	for k=1 to mesh.numfaces do
	(
		matid = getfacematid mesh k
		if (matclass == "Multimaterial") then
		(
			local subid = (mod matid mesh.material.count) as integer
			if (subid == 0) then subid = mesh.material.count
			matid = subid
		)
		else 
		(
			matid = 1
		)
			
		j=1 
		local found = false
		
		while (j<mats.count and found == false) do
		(
			if mats[j] == matid then found = true
			else j+=1
		)
		
		M3E_buildFaces mesh k div.regn[j]
	)
	
	return div
)

fn M3E_buildVertices verts =
(
	for i = 1 to verts.count do
	(
		local v = verts[i]
		local cnorm, cuv
			
		-- convert from float to byte
		for i = 1 to 4 do
		(
			v.bw[i] = (v.bw[i] * 255) as integer
		)
		
		-- compress normals
		cnorm = #(0,0,0,0)
		for i = 1 to 3 do
		(
			local cn = 1 + v.normal[i]
			cn = (cn / 2) * 255
			cnorm[i] = cn as integer
		)
		v.normal = cnorm
		
		-- compress UV's
		cuv = #(0,0)
		if (g_flipuvy) then v.uv[2] = 1 - v.uv[2]
		for i = 1 to 2 do
		(
			
			cuv[i] = (v.uv[i] * 2046) as integer
		)
		
		v.uv = cuv
		
		verts[i] = v
	)
)

fn M3E_getVertices m3obj =
(
	echo "Building vertex list..."
	local varray = #()
	
	if (m3obj.skin != undefined) then
	(
		modPanel.setCurrentObject m3obj.skin
	)
	
	local regn = m3obj.submeshes.regn
	
	for i = 1 to regn.count do
	(
		sm = regn[i]
		
		for j = 1 to sm.verts.count do
		(
			local v = sm.verts[j]
			local vIndex = v.origvertindx
			
			-- set default incase no skin is present
			v.bw 	= #(1,0,0,0)
			v.bi	= #(0,0,0,0)			
			
			if (m3obj.skin != undefined) then
			(
				nWeights = skinops.getvertexweightcount m3obj.skin vIndex
				for w = 1 to nWeights do
				(
					weightval = skinops.getvertexweight m3obj.skin vIndex w
					vboneid = skinops.getvertexweightboneid m3obj.skin vIndex w
					local skinBoneid = (findItem sm.skinBones vboneid) - 1
					
					if (skinBoneid == -1) then 
					(
						-- create local list
						append sm.skinBones vboneid
						-- create global list
						append m3obj.skinBones vboneid
						-- adjust boneid to be based on local list for verts
						skinBoneid = (sm.skinBones.count - 1)
					)
					
					v.bw[w] = weightval
					v.bi[w] = skinBoneid
				)
			)
			
			append varray v
		)
		
		regn[i] = sm
	)
	
	-- build vertices for export
	M3E_buildVertices varray

	return varray
)

fn M3E_getSkinnedBones m3obj =
(
	echo "Getting skinned bones..."
	-- get proper skinned bones
	local m3skinBones = #()
	if (m3obj.skin != undefined) then
	(
		m3skinBones[m3obj.skinBones.count] = 0
		max modify mode
		modPanel.setCurrentObject m3obj.skin
		
		for i = 1 to m3obj.skinBones.count do
		(
			b = m3obj.skinBones[i]
			sbName = skinops.getbonename m3obj.skin b 0
			for j = 1 to m3obj.bones.count do
			(
				m3bone = m3obj.bones[j]
				if (sbName == m3bone.name) then
				(
					m3skinBones[i] = (j - 1)
				)
			)
		)
	)
	else
	(
		-- No skin present, static single skinned bone list
		append m3skinBones 0
	)
	
	return m3skinBones
)

fn M3E_checkPath cPath =
(
	local modPath = cPath
	-- check for type of path 
	-- hard disk 
	if (matchPattern modPath pattern:"*:\\*") then
	(
		-- check if hard disk path has end backslash, add it if not
		if (modPath[modPath.count] != "\\") then modPath += "\\"
	)
	else
	(
		-- internal MPQ
		if (modPath[modPath.count] != "/") then modPath += "/"
	)
	
	return modPath
)

fn M3E_getMaterials m3obj =
(
	echo "Getting materials..."
	
	mtlarray = #()
	mesh = m3obj.mesh
	mats = m3obj.usedMats
	local parentmatclass = classOf mesh.material as string

	for h = 1 to mats.count do
	(
		mtl = M3E_Mat()
		matid = mats[h]
		
		if (parentmatclass == "Multimaterial") then
		(
			mmat = mesh.material[matid]
		)
		else
		(
			mmat = mesh.material
		)
		
		mtl.name = mmat.name

		local matclass = classOf mmat as string
		
		if (matclass != "Standardmaterial" and matclass != "sc2material") then
		(
			if (parentmatclass == "Multimaterial") then
			(
				messagebox "Multimaterial submaterials must be of Standard or Starcraft 2 type!"
				throw "Multimaterial submaterials must be of Standard or Starcraft 2 type"
			)
			else
			(
				messagebox "Materials must be of Standard or Starcraft 2 type!"
				throw "Materials must be of Standard or Starcraft 2 type"				
			)
		)
		
		-- Export UI settings
		-- 1. Standard Material
		if (matclass == "Standardmaterial") then
		(
			mtl.spec	= mmat.specularLevel
		)
		
		-- 2. Starcraft 2 Material
		if (matclass == "sc2material") then
		(
			-- Flag set
			local flags = 0
			--{
			flags = bit.set flags 3 mmat.Unfogged
			flags = bit.set flags 4 mmat.TwoSided
			flags = bit.set flags 5 mmat.Unshaded
			flags = bit.set flags 6 mmat.NoShadowsCast
			flags = bit.set flags 7 mmat.NoHitTest
			flags = bit.set flags 8 mmat.NoShadowsReceived
			flags = bit.set flags 9 mmat.DepthPrepass
			flags = bit.set flags 10 mmat.UseTerrainHDR
			flags = bit.set flags 12 mmat.SplatUVfix
			flags = bit.set flags 13 mmat.SoftBlending
			--}
			
			mtl.flags				= flags
			mtl.blendMode 	= (mmat.blendMode - 1) -- -1 due to using 1 based index
			mtl.priority			= mmat.priority
			mtl.spec				= mmat.specVal
			mtl.spec_mult		= mmat.specMult
			mtl.emis_mult		= mmat.emisMult
			mtl.cutoutThresh = mmat.cutoutThresh
			mtl.layerBlendType = (mmat.layerBlendType - 1)
			mtl.emisBlendType = (mmat.emissiveBlendType - 1)
			
			if (mmat.alphaMaskMap != undefined) then mtl.alphaMaskEnable = 1 else mtl.alphaMaskEnable = 0
		)
		
		-- Map handling
		for k = 1 to 13 do
		(
			local layr = M3E_LAYR()
			local fullPath = ""
			local layrBitmap = ""
			local curMap

			-- Different Material Handling
			-- 1. Standard Material
			if (matclass == "Standardmaterial") then
			(
				case k of
				(
					1: curMap = mmat.diffuseMap
					3: curMap = mmat.specularMap
					4: curMap = mmat.selfIllumMap
					6: curMap = mmat.reflectionMap
					10: curMap = mmat.bumpMap
				)
			)
			
			-- 2. Starcraft 2 Material
			if (matclass == "sc2material") then
			(
				case k of
				(
					1: curMap = mmat.diffuseMap
					2: curMap = mmat.decalMap
					3: curMap = mmat.specularMap
					4: curMap = mmat.selfIllumMap
					6: curMap = mmat.envioMap
					7: curMap = mmat.envioMaskMap
					8: curMap = mmat.alphaMaskMap
					10: curMap = mmat.bumpMap
					11: curMap = mmat.heightMap
				)
			)
			
			if (curMap != undefined) then
			(
				local mapClass = classOf curMap as string
				
				if (mapClass == "sc2bitmap") then
				(
					layrBitmap = curMap.scbitmap.filename
					
					local aflags = 0
					aflags = bit.set aflags 1 curMap.TeamColour
					aflags = bit.set aflags 2 curMap.TexAlphaOnly
					layr.alphaFlags = aflags
					
					layr.flags = bit.set layr.flags 3 curMap.TexWrap
				)
				else if (mapClass == "Bitmaptexture") then
				(
					layrBitmap = curMap.bitmap.filename
				)
				else
				(
					messageBox "Map textures must be of Bitmap or Starcraft 2 Bitmap type!"
					throw "Map textures must be of Bitmap or Starcraft 2 Bitmap type"
				)
			)
			
			if (g_UseInternalPath == false) then
			(
				fullPath = layrBitmap
			)
			else
			(
				if (layrBitmap != "") then
				(
					fullPath = (M3E_checkPath texPath) + (M3E_Name layrBitmap "\\")
				)
				else
				(
					fullPath = ""
				)
			)

			layr.name = fullPath
			append mtl.layrs #(layr)
		)

		append mtlarray mtl
	)
	
	--echo ("nMaterials: "+mtlarray.count as string)
	return mtlarray
)

---------------------
-- M3 HELPERS  --
---------------------

fn WriteRef bitStream data reftype subtype:#none flags:0  =
(
	local ref = M3E_Ref()
	-- Make sure it isn't a null reference...
	if (data != #null and data != undefined) then
	(
		-- MODL bookmark
		local bm = ftell bitStream
		
		-- head to end of file to write reference data
		fseek bitStream 0 #seek_end
		
		-- Calculate chunk size
		-- need special VERT clause
		local byte_size = (M3E_getSize reftype) * data.count
		case reftype of
		(
			#VERT: (M3E_Tag.TagAppend bitStream #U8__ (byte_size) twrite flags:flags; ref.entries = byte_size;)
			#CHAR: (byte_size += 1; M3E_Tag.TagAppend bitStream #CHAR (byte_size) twrite flags:flags)
			default: M3E_Tag.TagAppend bitStream reftype (data.count) twrite flags:flags
		)
		WriteBlank bitStream byte_size flags:#none ret:true
		ref.refid = twrite.count - 1 -- convert to base 0
		
		-- ** REF DEBUG **
		/* 
		echo ("reftype: "+reftype as string)
		echo ("refid: "+ref.refid as string)
		echo ("data.count: "+data.count as string)
		echo ("byte_size: "+byte_size as string)
		echo ("ftell: "+(ftell bitstream) as string)
		*/
		
		if (reftype != #CHAR) then
		(
			for i = 1 to data.count do
			(
				d = data[i]
				case reftype of
				(
					#SEQS: M3E_SEQS.Write bitStream d
					#STC_: M3E_STC.Write bitStream d
					#SDEV: M3E_AnimBlock.Write bitStream d #EVNT
					#EVNT: M3E_Event.Write bitStream d
					#STG_: M3E_STG.Write bitStream d
					#STS_: M3E_STS.Write bitStream d
					#BONE: M3E_Bone.Write bitStream d
					#VERT: M3E_Vertex.Write bitStream d
					#REGN: M3E_REGN.Write bitStream d
					#BAT_: M3E_BAT.Write bitStream d
					#MSEC: M3E_MSEC.Write bitStream d
					#ATT_: M3E_Attachment.Write bitStream d
					#MATM: M3E_MATM.Write bitStream d
					#MAT_: M3E_MAT.Write bitStream d
					#LAYR: M3E_LAYR.Write bitStream d
					#DIV_: M3E_DIV.Write bitStream d
					#IREF: WriteVec bitStream d #matrix
					#U16_: WriteArray bitStream d flags:subtype abool:false
					#U32_: WriteArray bitStream d flags:subtype abool:false
					#I32_: WriteArray bitStream d flags:subtype abool:false
				)
			)
		)
		else
		(
			WriteString bitStream data
			-- fix tag entry
			twrite[twrite.count].nData = byte_size
			ref.entries = byte_size
		)
		
		if (ref.entries == undefined) then ref.entries = data.count
		-- return to MODL bookmark
		fseek bitStream bm #seek_set
	)
	else
	(
		-- setup blank reference
		ref.entries = 0
		ref.refid = 0
	)
	
	-- write reference 
	M3E_Ref.Write bitStream ref
)

fn WriteArray bitStream data flags:#uint32 reftype:#null abool:true  =
(
	-- if not array...
	if (abool != true) then icount = 1 else icount = data.count
	for i = 1 to icount do
	(
		local d
		if (abool != true) then d = data else d = data[i]
		case flags of
		(
			#Ref: WriteRef bitStream d reftype
			#uint32: writeLong bitStream d #unsigned
			#int32: WriteLong bitStream d #signed
			#uint16: writeShort bitStream d #unsigned
			#int16: writeShort bitStream d #signed
			#uint8: writeByte bitStream d #unsigned
			#int8: writeByte bitStream d #signed
			#float: writeFloat bitStream d
			#v3d: WriteVec bitStream d #v3d
			#animind: M3E_AnimOffs.Write bitStream d
		)
	)
)

fn M3E_GetMODL m3obj =
(
	m3 = M3E_MODL()
	
	--m.name = "Test.max"
	m3.name = m3obj.mesh.name
	m3.versID = 1576275
	m3.SEQS = #(M3E_SEQS.Default(m3obj))
	m3.STC = #(M3E_STC.Default())
	m3.STG = #(M3E_STG())
	m3.vec1 = [0,0,0]
	m3.STS = #(M3E_STS.Default m3)
	m3.bones = m3obj.bones

	local nSkinBones = 0
	for i = 1 to m3obj.skinBones.count do
	(
		local skinBoneInd = m3obj.skinBones[i] + 1
		if (skinBoneInd > nSkinBones) then nSkinBones = m3obj.skinBones[i] + 1
	)
	m3.nSkinBones = nSkinBones

	-- grab submeshes first
	m3.DIV = #(m3obj.submeshes)
	
	-- build verts based on submesh list
	m3.vflags = 25297021
	m3.verts = m3obj.verts
	m3.bonelu = m3obj.skinBones
	m3.ext1 = M3E_Sphere.GetBound m3obj.mesh
	if (m3obj.attachments != undefined) then
	(
		m3.attach = m3obj.attachments
		m3.attachlu = #()
		for i = 1 to m3obj.attachments.count do append m3.attachlu -1
	)
	m3.matm = M3E_MATM.Read m3obj.usedMats
	m3.mat = m3obj.materials
	m3.iref = #()
	for i = 1 to m3.bones.count do
	(
		append m3.iref m3.bones[i].iref
	)

	return m3
)

fn M3E_getScene =
(
	step = "Gathering Scene Objects"
	echo "Gathering scene objects..."

	m3model = M3E_Model()
	m3model.mesh = M3E_getMesh()
	m3model.skin	= M3E_getSkin m3model.mesh
	m3model.bones = M3E_getBones m3model.skin
	m3model.usedMats = M3E_getUsedMats m3model.mesh
	m3model.submeshes = M3E_getSubmesh m3model
	m3model.verts = M3E_getVertices m3model
	M3E_buildDIV m3model 	-- Build REGN's for model DIV
	m3model.skinBones = M3E_getSkinnedBones m3model
	m3model.materials = M3E_getMaterials m3model
	if m3model.skin != undefined then
	(
		local objAttachs = M3E_getObjects helpers "Point"
		m3model.attachments = M3E_buildAttachments objAttachs m3model.bones -- Build attachments
	)
	else m3model.attachments = #()
	
	echo "Gathering scene data done!"
)

fn M3E_buildM3 m3obj =
(
	step = "Build M3"
	format "Building M3 structures..." to:listener
	head = M3E_Header.Default()
	-- Gather MODL data
	-- assign global, will update this structure throughout export
	mwrite = M3E_GetMODL m3obj
	
	echo "done!"
)

fn M3E_Start_Write bitStream =
(
	step = "Start Model Write"

	-- write Header and blank bytes and tag reference
	echo "Writing buffers..."
	M3E_Tag.TagAppend bitStream #MD34 1 twrite flags:11
	WriteBlank bstream (M3E_getSize #MD34) flags:#none
	
	-- Create MODL tag, write MODL blank bytes
	-- using model version 23
	M3E_Tag.TagAppend bitStream #MODL 1 twrite flags:23	
	WriteBlank bstream (M3E_getSize #MODL flags:23) flags:#none
	
	echo "Writing model data..."
	-- Write Model Data first
	M3E_MODL.Write bitStream mwrite 23	

	echo "Writing tag data..."
	-- Write Tags second...
	M3E_Tag.Write bitStream twrite
	
	echo "Writing header data..."
	-- ... and write Header last
	M3E_Header.Write bitStream head
	echo "Model writing done!\n"
)

fn M3E_Main file =
(
	try 
	(
		echo "\n====== Export Starting ======\n"
		M3E_Reset_Globals()
		scrStart = TimeStamp() -- start tracking script processing time
		M3E_getScene()
		M3E_buildM3(m3model)
	
		echo "\n==== Model Information ===="
		echo ("Mesh Name: "+m3model.mesh.name)
		echo ("Vertices: "+m3model.verts.count as string)
		echo ("Faces: "+m3model.mesh.numfaces as string)
		echo ("Submeshes: "+m3model.submeshes.regn.count as string)
		echo ("Materials: "+m3model.materials.count as string)
		echo ("Bones: "+m3model.bones.count as string)
		echo ("Skinned bones: "+m3model.skinBones.count as string)
		echo "" -- spacer
		
		bstream = M3E_Open file "wb"
		M3E_Start_Write bstream
		M3E_Close bstream
		
		-- Calculate Script Execution time
		scrEnd = TimeStamp()
		format "Export took % seconds\n" ((scrEnd - scrStart) / 1000.0)
		echo "====== Export Successful ======\n"
	)
	catch
	(
		echo ("****"+getCurrentException()+"****")
		echo "====== Export Failed ======\n"
	)
	
	clearSelection()
	setCommandPanelTaskMode #utility
	--M3E_Reset_Globals()
)

-- *********************************
--  USER INTERFACE FUNCTIONS
-- *********************************
-- Glob UI Funcs
fn M3E_GetFile ftypes =
(
	local ret = getSaveFileName types:ftypes
	if ret != undefined then return ret
)

utility M3E_export "M3 - Export"
(
	group "Texture Options"
	(
		label	lblTexPath "Texture Path:" align:#left
		edittext tTexPath text:texPath
		button bBrowsePath "Browse"
		checkbox chkUseInternalPath "Use Internal Path" checked:(g_UseInternalPath)
		checkbox chkFlipUVY "Flip UV Y-coord" checked:(g_flipuvy)
	)
	
	group "Animation Options"
	(
		label lblbindFrames "Pose Frames:" align:#left
		spinner spnBindposeFrame "Bindpose:" range:[0,1e8,0] type:#integer
		spinner spnBaseFrame "Base:" range:[0,1e8,1] type:#integer
	)
	
	button bExport "Export" height:35 width:100 offset:[0,5]
	button bAbout "About" offset:[0,5]
	
	on M3E_export open do
	(
		if (chkUseInternalPath.checked == true) then
		(
			tTexPath.enabled = true
			bBrowsePath.enabled = true
		)
		else
		(
			tTexPath.enabled = false
			bBrowsePath.enabled = false			
		)
	)
	on bBrowsePath pressed do
	(
		local tPath = getSavePath caption:"Texture Path"
		if (tPath != undefined) then tTexPath.text = tPath
	)
	on chkUseInternalPath changed state do
	(
		if state == on then
		(
			tTexPath.enabled = true
			bBrowsePath.enabled = true
		)
		else
		(
			tTexPath.enabled = false
			bBrowsePath.enabled = false
		)
	)
	on bExport pressed do
	(
		local fname = getSaveFileName types:"M3 model (*.m3)|*.m3|All Files|*.*|"
		if fname != undefined then
		(
			g_UseInternalPath = chkUseInternalPath.checked
			g_flipuvy = chkFlipUVY.checked
			texPath = tTexPath.text
			g_bindposeFrame = spnBindposeFrame.value
			g_baseFrame = spnBaseFrame.value
			M3E_Main(fname)
		)
	)
	on bAbout pressed do
	(
		messagebox "M3 Exporter v0.14 - 28-05-2010\n\xa9 2010, NiNtoxicated (madyavic@gmail.com)\nVisit www.sc2mapster.com for updates and more information\nHead to the forum for support. Feedback is appreciated!" title:"About"
	)
)